﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Clients.Hive.ExperimentManager;
using HeuristicLab.Clients.Hive.Jobs;
using HeuristicLab.Hive;

namespace HeuristicLab.Clients.Hive {
  public class HiveJobDownloader {
    private IEnumerable<Guid> jobIds;
    private JobDownloader<ItemJob> jobDownloader;
    private IDictionary<Guid, HiveJob> results;

    public bool IsFinished {
      get {
        //return tasks.TrueForAll(t => t.Status == TaskStatus.RanToCompletion ||
        //                             t.Status == TaskStatus.Faulted ||
        //                             t.Status == TaskStatus.Canceled);
        return results.Count == jobIds.Count();
      }
    }

    public int FinishedCount {
      get {
        //var faulted = tasks.Where(t => t.Status == TaskStatus.Faulted);
        //if (faulted.Count() > 0) {
        //  abort = true;
        //  throw faulted.First().Exception;
        //}
        //return tasks.Count(t => t.Status == TaskStatus.RanToCompletion ||
        //                        t.Status == TaskStatus.Faulted ||
        //                        t.Status == TaskStatus.Canceled);
        return results.Count;
      }
    }

    public IDictionary<Guid, HiveJob> Results {
      get {
        //var results = new Dictionary<Guid, HiveJob>();
        //foreach (var t in tasks) {
        //  if (t.Status == TaskStatus.Faulted) {
        //    throw t.Exception;
        //  }
        //  if (t.Result != null)
        //    results.Add(t.Result.Job.Id, t.Result);
        //}
        return results;
      }
    }

    public HiveJobDownloader(IEnumerable<Guid> jobIds) {
      this.jobIds = jobIds;
      this.jobDownloader = new JobDownloader<ItemJob>(2, 2);
      this.results = new Dictionary<Guid, HiveJob>();
    }

    public void StartAsync() {
      foreach (Guid jobId in jobIds) {
        jobDownloader.DownloadJob(jobId,
          (id, itemJob, exception) => {
            if (exception != null) {
              throw new JobDownloaderException("Downloading job failed", exception);
            }
            Job job = ServiceLocator.Instance.CallHiveService(s => s.GetJob(id));
            if (job != null && itemJob != null) {
              HiveJob hiveJob;
              if (itemJob is OptimizerJob) {
                hiveJob = new OptimizerHiveJob((OptimizerJob)itemJob);
              } else {
                hiveJob = new HiveJob(itemJob, true);
              }
              hiveJob.Job = job;
              this.results.Add(id, hiveJob);
            }
          });
      }

      //tasks = new List<Task<HiveJob>>();
      //TaskScheduler.UnobservedTaskException += new EventHandler<UnobservedTaskExceptionEventArgs>(TaskScheduler_UnobservedTaskException);
      //foreach (Guid jobId in jobIds) {
      //  tasks.Add(Task<JobData>.Factory.StartNew(
      //    (x) => DownloadJob(x), jobId)
      //    .ContinueWith((x) => DeserializeJob(x.Result)));
      //}
    }

    //private void TaskScheduler_UnobservedTaskException(object sender, UnobservedTaskExceptionEventArgs e) {
    //  e.SetObserved(); // avoid crash of process because task crashes. first exception found is handled in Results property
    //}

    //// use semaphore to ensure only few concurrenct connections and few SerializedJob objects in memory
    //private Semaphore downloadSemaphore = new Semaphore(2, 2);
    //private Semaphore deserializeSemaphore = new Semaphore(2, 2);
    //protected JobData DownloadJob(object jobId) {
    //  downloadSemaphore.WaitOne();
    //  deserializeSemaphore.WaitOne();
    //  JobData result;
    //  try {
    //    if (abort) return null;
    //    result = ServiceLocator.Instance.CallHiveService(s => s.GetJobData((Guid)jobId));
    //  }
    //  finally {
    //    downloadSemaphore.Release();
    //  }
    //  return result;
    //}

    //protected HiveJob DeserializeJob(JobData jobData) {
    //  try {
    //    Job job = ServiceLocator.Instance.CallHiveService(s => s.GetJob(jobData.JobId));
    //    if (abort || job == null || jobData == null) return null;

    //    HiveJob hiveJob;
    //    var itemJob = PersistenceUtil.Deserialize<ItemJob>(jobData.Data);
    //    if (itemJob is OptimizerJob) {
    //      hiveJob = new OptimizerHiveJob((OptimizerJob)itemJob);
    //    } else {
    //      hiveJob = new HiveJob(itemJob, true);
    //    }
    //    jobData.Data = null; // reduce memory consumption.
    //    hiveJob.Job = job;
    //    return hiveJob;
    //  }
    //  finally {
    //    deserializeSemaphore.Release();
    //  }
    //}
  }
}
